# Copyright 2004-2005 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
from twisted.internet import task
if utils.checkTwisted():
	from twisted.xish.domish import Element
else:
	from tlib.domish import Element
from tlib import msn, msnp2p
import sha

import groupchat
import avatar
import msnw
import config
import debug
import lang




name = "MSN Transport"   # The name of the transport
url = "http://msn-transport.jabberstudio.org"
version = "0.10"         # The transport version
mangle = True            # XDB '@' -> '%' mangling
id = "msn"               # The transport identifier


# Load the default avatar
f = open("legacy/defaultJabberAvatar.png")
defaultJabberAvatarData = f.read()
f.close()


def isGroupJID(jid):
	""" Returns True if the JID passed is a valid groupchat JID (for MSN, does not contain '%') """
	return (jid.find('%') == -1)


	
# This should be set to the name space the registration entries are in, in the xdb spool
namespace = "jabber:iq:register"


def formRegEntry(username, password):
	""" Returns a domish.Element representation of the data passed. This element will be written to the XDB spool file """
	reginfo = Element((None, "query"))
	reginfo.attributes["xmlns"] = "jabber:iq:register"
	
	userEl = reginfo.addElement("username")
	userEl.addContent(username)
	
	passEl = reginfo.addElement("password")
	passEl.addContent(password)

	return reginfo




def getAttributes(base):
	""" This function should, given a spool domish.Element, pull the username, password,
	and out of it and return them """
	username = ""
	password = ""
	for child in base.elements():
		try:
			if child.name == "username":
				username = child.__str__()
			elif child.name == "password":
				password = child.__str__()
		except AttributeError:
			continue
	
	return username, password


def startStats(statistics):
	stats = statistics.stats
	stats["MessageCount"] = 0
	stats["FailedMessageCount"] = 0
	stats["AvatarCount"] = 0
	stats["FailedAvatarCount"] = 0

def updateStats(statistics):
	stats = statistics.stats
	stats["AvatarCount"] = msnp2p.MSNP2P_Avatar.TRANSFER_COUNT
	stats["FailedAvatarCount"] = msnp2p.MSNP2P_Avatar.ERROR_COUNT


def msn2jid(msnid):
	""" Converts a MSN passport into a JID representation to be used with the transport """
	return msnid.replace('@', '%') + "@" + config.jid

translateAccount = msn2jid # Marks this as the function to be used in jabber:iq:gateway (Service ID Translation)

def jid2msn(jid):
	""" Converts a JID representation of a MSN passport into the original MSN passport """
	return unicode(jid[:jid.find('@')].replace('%', '@'))


def presence2state(show, ptype): 
	""" Converts a Jabber presence into an MSN status code """
	if ptype == "unavailable":
		return msn.STATUS_OFFLINE
	elif not show or show == "online" or show == "chat":
		return msn.STATUS_ONLINE
	elif show == "dnd":
		return msn.STATUS_BUSY
	elif show == "away" or show == "xa":
		return msn.STATUS_AWAY


def state2presence(state):
	""" Converts a MSN status code into a Jabber presence """
	if state == msn.STATUS_ONLINE:
		return (None, None)
	elif state == msn.STATUS_BUSY:
		return ("dnd", None)
	elif state == msn.STATUS_AWAY:
		return ("away", None)
	elif state == msn.STATUS_IDLE:
		return ("away", None)
	elif state == msn.STATUS_BRB:
		return ("away", None)
	elif state == msn.STATUS_PHONE:
		return ("dnd", None)
	elif state == msn.STATUS_LUNCH:
		return ("away", None)
	else:
		return (None, "unavailable")




# This class handles groupchats with the legacy protocol
class LegacyGroupchat(groupchat.BaseGroupchat):
	def __init__(self, session, resource, ID=None, existing=False, switchboardSession=None):
		""" Possible entry points for groupchat
			- User starts an empty switchboard session by sending presence to a blank room
			- An existing switchboard session is joined by another MSN user
			- User invited to an existing switchboard session with more than one user
		"""
		groupchat.BaseGroupchat.__init__(self, session, resource, ID)
		if not existing:
			self.switchboardSession = msnw.GroupchatSwitchboardSession(self, makeSwitchboard=True)
		else:
			self.switchboardSession = switchboardSession
			
		assert(self.switchboardSession != None)
		
		debug.log("LegacyGroupchat: \"%s\" created" % (self.roomJID()))
	
	def removeMe(self):
		self.switchboardSession.removeMe()
		self.switchboardSession = None
		groupchat.BaseGroupchat.removeMe(self)
		debug.log("LegacyGroupchat: \"%s\" destroyed" % (self.roomJID()))
		utils.mutilateMe(self)
	
	def sendLegacyMessage(self, message, noerror):
		debug.log("LegacyGroupchat: \"%s\" sendLegacyMessage(\"%s\")" % (self.roomJID(), message))
		self.switchboardSession.sendMessage(message, noerror)
	
	def sendContactInvite(self, contactJID):
		debug.log("LegacyGroupchat: \"%s\" sendContactInvite(\"%s\")" % (self.roomJID(), contactJID))
		userHandle = jid2msn(contactJID)
		self.switchboardSession.inviteUser(userHandle)



# This class handles most interaction with the legacy protocol
class LegacyConnection(msnw.MSNConnection):
	""" A glue class that connects to the legacy network """
	def __init__(self, username, password, session):
		self.session = session
		self.listSynced = False
		self.initialListVersion = 0

		self.remoteShow = ""
		self.remoteStatus = ""
		self.remoteNick = ""

		# Init the MSN bits
		msnw.MSNConnection.__init__(self, username, password)

		# User typing notification stuff
		self.userTyping = dict() # Indexed by contact MSN ID, stores whether the user is typing to this contact
		# Contact typing notification stuff
		self.contactTyping = dict() # Indexed by contact MSN ID, stores an integer that is incremented at 5 second intervals. If it reaches 3 then the contact has stopped typing. It is set to zero whenever MSN typing notification messages are received
		# Looping function
		self.userTypingSend = task.LoopingCall(self.sendTypingNotifications)
		self.userTypingSend.start(5.0)
		
		import legacylist # Is in here to prevent an ImportError loop
		self.legacyList = legacylist.LegacyList(self.session)
	
		debug.log("LegacyConnection: \"%s\" - created" % (self.session.jabberID))
	
	def removeMe(self):
		debug.log("LegacyConnection: \"%s\" - being deleted" % (self.session.jabberID))
	
		self.userTypingSend.stop()
	
		msnw.MSNConnection.removeMe(self)
		self.legacyList.removeMe()
		self.legacyList = None
		self.session = None

		utils.mutilateMe(self)
	
	def jidRes(self, resource):
		to = self.session.jabberID
		if resource:
			to += "/" + resource

		return to
	
	def highestResource(self):
		""" Returns highest priority resource """
		return self.session.highestResource()
	
	def sendMessage(self, dest, resource, body, noerror):
		dest = jid2msn(dest)
		if self.userTyping.has_key(dest):
			del self.userTyping[dest]
		try:
			msnw.MSNConnection.sendMessage(self, dest, resource, body, noerror)
			self.session.pytrans.statistics.stats["MessageCount"] += 1
		except:
			self.failedMessage(dest, body)
			raise
	
	def msnAlert(self, text, actionurl, subscrurl):
		if not self.session: return

		el = Element((None, "message"))
		el.attributes["to"] = self.session.jabberID
		el.attributes["from"] = config.jid
		el.attributes["type"] = "headline"
		body = el.addElement("body")
		body.addContent(text)
		
		x = el.addElement("x")
		x.attributes["xmlns"] = "jabber:x:oob"
		x.addElement("desc").addContent("More information on this notice.")
		x.addElement("url").addContent(actionurl)

		x = el.addElement("x")
		x.attributes["xmlns"] = "jabber:x:oob"
		x.addElement("desc").addContent("Manage subscriptions to alerts.")
		x.addElement("url").addContent(subscrurl)

		self.session.pytrans.send(el)
	
	def setStatus(self, nickname, show, status):
		statusCode = presence2state(show, None)
		msnw.MSNConnection.changeStatus(self, statusCode, nickname, status)
	
	def updateAvatar(self, av=None):
		global defaultJabberAvatarData

		if av:
			msnw.MSNConnection.changeAvatar(self, av.getImageData())
		else:
			msnw.MSNConnection.changeAvatar(self, defaultJabberAvatarData)
	
	def sendTypingNotifications(self):
		if not self.session: return
	
		# Send any typing notification messages to the user's contacts
		for contact in self.userTyping.keys():
			if self.userTyping[contact]:
				self.sendTypingToContact(contact)

		# Send any typing notification messages from contacts to the user
		for contact, resource in self.contactTyping.keys():
			self.contactTyping[(contact, resource)] += 1
			if self.contactTyping[(contact, resource)] >= 3:
				self.session.sendTypingNotification(self.jidRes(resource), msn2jid(contact), False)
				del self.contactTyping[(contact, resource)]
	
	def gotContactTyping(self, contact, resource):
		if not self.session: return
		# Check if the contact has only just started typing
		if not self.contactTyping.has_key((contact, resource)):
			self.session.sendTypingNotification(self.jidRes(resource), msn2jid(contact), True)

		# Reset the counter
		self.contactTyping[(contact, resource)] = 0
	
	def userTypingNotification(self, dest, resource, composing):
		if not self.session: return
		dest = jid2msn(dest)
		self.userTyping[dest] = composing
		if composing: # Make it instant
			self.sendTypingToContact(dest)
	
	def listSynchronized(self):
		if not self.session: return
		self.session.sendPresence(to=self.session.jabberID, fro=config.jid)
		self.legacyList.syncJabberLegacyLists()
		self.listSynced = True
		#self.legacyList.flushSubscriptionBuffer()
	
	def gotMessage(self, remoteUser, resource, text):
		if not self.session: return
		source = msn2jid(remoteUser)
		self.session.sendMessage(self.jidRes(resource), fro=source, body=text, mtype="chat")
		self.session.pytrans.statistics.stats["MessageCount"] += 1
	
	def avatarHashChanged(self, userHandle, hash):
		if not self.session: return
		av = self.session.pytrans.avatarCache.getAvatar(hash)
		if av:
			msnContact = self.getContacts().getContact(userHandle)
			msnContact.msnobjGot = True
			jid = msn2jid(userHandle)
			c = self.session.contactList.findContact(jid)
			if not c: return
			c.updateAvatar(av)
		else:
			self.requestAvatar(userHandle)
	
	def gotAvatarImage(self, userHandle, imageData):
		if not self.session: return
		jid = msn2jid(userHandle)
		c = self.session.contactList.findContact(jid)
		if not c: return
		av = self.session.pytrans.avatarCache.setAvatar(imageData)
		c.updateAvatar(av)
	
	def loggedIn(self):
		if not self.session: return
		debug.log("LegacyConnection: \"%s\" - loggedIn()" % (self.session.jabberID))
		self.session.ready = True
	
	def contactStatusChanged(self, remoteUser):
		if not (self.session and self.getContacts()): return
		debug.log("LegacyConnection: \"%s\" - contactStatusChanged(\"%s\")" % (self.session.jabberID, remoteUser))
		
		msnContact = self.getContacts().getContact(remoteUser)
		c = self.session.contactList.findContact(msn2jid(remoteUser))
		if not (c and msnContact): return

		show, ptype = state2presence(msnContact.status)
		status = msnContact.personal.decode("utf-8")
		screenName = msnContact.screenName.decode("utf-8")

		c.updateNickname(screenName, push=False)
		c.updatePresence(show, status, ptype)
	
	def ourStatusChanged(self, statusCode):
		# Send out a new presence packet to the Jabber user so that the MSN-t icon changes
		if not self.session: return
		debug.log("LegacyConnection: \"%s\" - ourStatusChanged(\"%s\")" % (self.session.jabberID, statusCode))
		self.remoteShow, ptype = state2presence(statusCode)
		self.sendShowStatus()
	
	def ourPersonalChanged(self, statusMessage):
		if not self.session: return
		debug.log("LegacyConnection: \"%s\" - ourPersonalChanged(\"%s\")" % (self.session.jabberID, statusMessage))
		self.remoteStatus = statusMessage
		self.sendShowStatus()
	
	def ourNickChanged(self, nick):
		if not self.session: return
		debug.log("LegacyConnection: \"%s\" - ourNickChanged(\"%s\")" % (self.session.jabberID, nick))
		self.remoteNick = nick
		self.sendShowStatus()
	
	def sendShowStatus(self):
		if not self.session: return
		source = config.jid
		to = self.session.jabberID
		self.session.sendPresence(to=to, fro=source, show=self.remoteShow, status=self.remoteStatus, nickname=self.remoteNick)
	
	def userMapping(self, passport, jid):
		if not self.session: return
		text = lang.get(self.session.lang).userMapping % (passport, jid)
		self.session.sendMessage(to=self.session.jabberID, fro=msn2jid(passport), body=text)
	
	def userAddedMe(self, userHandle):
		if not self.session: return
		self.session.contactList.getContact(msn2jid(userHandle)).contactRequestsAuth()
	
	def userRemovedMe(self, userHandle):
		if not self.session: return
		c = self.session.contactList.getContact(msn2jid(userHandle))
		c.contactDerequestsAuth()
		c.contactRemovesAuth()
	
	def serverGoingDown(self):
		if not self.session: return
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=lang.get(self.session.lang).msnMaintenance)
	
	def multipleLogin(self):
		if not self.session: return
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=lang.get(self.session.lang).msnMultipleLogin)
		self.session.removeMe()
	
	def accountNotVerified(self):
		if not self.session: return
		text = lang.get(self.session.lang).msnNotVerified % (self.session.username)
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text)
	
	def loginFailure(self, message):
		if not self.session: return
		text = lang.get(self.session.lang).msnLoginFailure % (self.session.username)
		self.session.sendErrorMessage(to=self.session.jabberID, fro=config.jid, etype="auth", condition="not-authorized", explanation=text, body="Login Failure")
	
	def failedMessage(self, remoteUser, message):
		if not self.session: return
		self.session.pytrans.statistics.stats["FailedMessageCount"] += 1
		fro = msn2jid(remoteUser)
		self.session.sendErrorMessage(to=self.session.jabberID, fro=fro, etype="wait", condition="recipient-unavailable", explanation=lang.get(self.session.lang).msnFailedMessage, body=message)
	
	def initialEmailNotification(self, inboxunread, foldersunread):
		if not self.session: return
		text = lang.get(self.session.lang).msnInitialMail % (inboxunread, foldersunread)
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text, mtype="headline")
	
	def realtimeEmailNotification(self, mailfrom, fromaddr, subject):
		if not self.session: return
		text = lang.get(self.session.lang).msnRealtimeMail % (mailfrom, fromaddr, subject)
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text, mtype="headline")
		
	def connectionLost(self, reason):
		if not self.session: return
		debug.log("LegacyConnection: \"%s\" - connectionLost(\"%s\")" % (self.session.jabberID, reason))
		text = lang.get(self.session.lang).msnDisconnected % ("Error") # FIXME, a better error would be nice =P
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text)
		self.session.removeMe() # Tear down the session


